# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    This module copyright (C) 2014 Therp BV (<http://therp.nl>)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################

import types
import logging
from openerp import release
from openerp.osv.orm import TransientModel
from openerp.osv import fields
from openerp.openupgrade.openupgrade import table_exists
from openerp.tools import config, safe_eval

# A collection of functions used in
# openerp/modules/loading.py

logger = logging.getLogger('OpenUpgrade')


def add_module_dependencies(cr, module_list):
    """
    Select (new) dependencies from the modules in the list
    so that we can inject them into the graph at upgrade
    time. Used in the modified OpenUpgrade Server,
    not to be called from migration scripts

    Also take the OpenUpgrade configuration directives 'forced_deps'
    and 'autoinstall' into account. From any additional modules
    that these directives can add, the dependencies are added as
    well (but these directives are not checked for the occurrence
    of any of the dependencies).
    """
    if not module_list:
        return module_list

    forced_deps = safe_eval.safe_eval(
        config.get_misc(
            'openupgrade', 'forced_deps_' + release.version,
            config.get_misc('openupgrade', 'forced_deps', '{}')))

    autoinstall = safe_eval.safe_eval(
        config.get_misc(
            'openupgrade', 'autoinstall_' + release.version,
            config.get_misc('openupgrade', 'autoinstall', '{}')))

    for module in list(module_list):
        module_list += forced_deps.get(module, [])
        module_list += autoinstall.get(module, [])

    module_list = list(set(module_list))

    dependencies = module_list
    while dependencies:
        cr.execute("""
            SELECT DISTINCT dep.name
            FROM
                ir_module_module,
                ir_module_module_dependency dep
            WHERE
                module_id = ir_module_module.id
                AND ir_module_module.name in %s
                AND dep.name not in %s
            """, (tuple(dependencies), tuple(module_list),))

        dependencies = [x[0] for x in cr.fetchall()]
        module_list += dependencies

    # Select auto_install modules of which all dependencies
    # are fulfilled based on the modules we know are to be
    # installed
    cr.execute("""
        SELECT name from ir_module_module WHERE state IN %s
        """, (('installed', 'to install', 'to upgrade'),))
    modules = list(set(module_list + [row[0] for row in cr.fetchall()]))
    cr.execute("""
        SELECT name from ir_module_module m
        WHERE auto_install IS TRUE
            AND state = 'uninstalled'
            AND NOT EXISTS(
                SELECT id FROM ir_module_module_dependency d
                WHERE d.module_id = m.id
                AND name NOT IN %s)
         """, (tuple(modules),))
    auto_modules = [row[0] for row in cr.fetchall()]
    if auto_modules:
        logger.info(
            "Selecting autoinstallable modules %s", ','.join(auto_modules))
        module_list += auto_modules

    return module_list


def log_model(model, local_registry):
    """
    OpenUpgrade: Store the characteristics of the BaseModel and its fields
    in the local registry, so that we can compare changes with the
    main registry
    """

    if not model._name:
        return

    # persistent models only
    if isinstance(model, TransientModel):
        return

    model_registry = local_registry.setdefault(
        model._name, {})
    if model._inherits:
        model_registry['_inherits'] = {'_inherits': unicode(model._inherits)}
    for k, v in model._columns.items():
        properties = {
            'type': v._type,
            'isfunction': (
                isinstance(v, fields.function) and 'function' or ''),
            'relation':
            v._type in ('many2many', 'many2one', 'one2many') and v._obj or '',
            'required': v.required and 'required' or '',
            'selection_keys': '',
            'req_default': '',
            'inherits': '',
            }
        if hasattr(v, 'oldname'):
            properties['oldname'] = v.oldname
        if v._type == 'selection':
            if hasattr(v.selection, "__iter__"):
                properties['selection_keys'] = unicode(
                    sorted([x[0] for x in v.selection]))
            else:
                properties['selection_keys'] = 'function'
        if v.required and k in model._defaults:
            if isinstance(model._defaults[k], types.FunctionType):
                # todo: in OpenERP 5 (and in 6 as well),
                # literals are wrapped in a lambda function
                properties['req_default'] = 'function'
            else:
                properties['req_default'] = unicode(
                    model._defaults[k])
        for key, value in properties.items():
            if value:
                model_registry.setdefault(k, {})[key] = value


def get_record_id(cr, module, model, field, mode):
    """
    OpenUpgrade: get or create the id from the record table matching
    the key parameter values
    """
    cr.execute(
        "SELECT id FROM openupgrade_record "
        "WHERE module = %s AND model = %s AND "
        "field = %s AND mode = %s AND type = %s",
        (module, model, field, mode, 'field')
        )
    record = cr.fetchone()
    if record:
        return record[0]
    cr.execute(
        "INSERT INTO openupgrade_record "
        "(module, model, field, mode, type) "
        "VALUES (%s, %s, %s, %s, %s)",
        (module, model, field, mode, 'field')
        )
    cr.execute(
        "SELECT id FROM openupgrade_record "
        "WHERE module = %s AND model = %s AND "
        "field = %s AND mode = %s AND type = %s",
        (module, model, field, mode, 'field')
        )
    return cr.fetchone()[0]


def compare_registries(cr, module, registry, local_registry):
    """
    OpenUpgrade: Compare the local registry with the global registry,
    log any differences and merge the local registry with
    the global one.
    """
    if not table_exists(cr, 'openupgrade_record'):
        return
    for model, flds in local_registry.items():
        registry.setdefault(model, {})
        for field, attributes in flds.items():
            old_field = registry[model].setdefault(field, {})
            mode = old_field and 'modify' or 'create'
            record_id = False
            for key, value in attributes.items():
                if key not in old_field or old_field[key] != value:
                    if not record_id:
                        record_id = get_record_id(
                            cr, module, model, field, mode)
                    cr.execute(
                        "SELECT id FROM openupgrade_attribute "
                        "WHERE name = %s AND value = %s AND "
                        "record_id = %s",
                        (key, value, record_id)
                        )
                    if not cr.fetchone():
                        cr.execute(
                            "INSERT INTO openupgrade_attribute "
                            "(name, value, record_id) VALUES (%s, %s, %s)",
                            (key, value, record_id)
                            )
                    old_field[key] = value



# -*- coding: utf-8 -*-
##############################################################################
#
#    OpenERP, Open Source Management Solution
#    This module copyright (C) 2014 Therp BV (<http://therp.nl>)
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Affero General Public License as
#    published by the Free Software Foundation, either version 3 of the
#    License, or (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Affero General Public License for more details.
#
#    You should have received a copy of the GNU Affero General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
##############################################################################
"""
This module can be used to contain functions that should be called at the end
of the migration. A migration may be run several times after corrections in
the code or the configuration, and there is no way for OpenERP to detect a
succesful result. Therefore, the functions in this module should be robust
against being run multiple times on the same database.
"""
import logging
from openerp import SUPERUSER_ID
from openerp.openupgrade import openupgrade
logger = logging.getLogger('OpenUpgrade.deferred')


def migrate_product_valuation(cr, pool):
    """Migrate the product valuation to a property field on product template.
    This field was moved to a new module which is not installed when the
    migration starts and thus not upgraded, which is why we do it here in the
    deferred step.

    This method removes the preserved legacy column upon success, to prevent
    double runs which would be harmful.
    """
    valuation_column = openupgrade.get_legacy_name('valuation')
    if not openupgrade.column_exists(cr, 'product_product', valuation_column):
        return

    cr.execute(
        """
        SELECT id FROM ir_model_fields
        WHERE model = 'product.template' AND name = 'valuation'
        """)
    field_id = cr.fetchone()[0]

    property_obj = pool['ir.property']
    default = 'manual_periodic'  # as per stock_account/stock_account_data.xml

    cr.execute(
        """
        SELECT product_tmpl_id, {column} FROM product_product
        WHERE {column} != %s""".format(column=valuation_column),
        (default,))
    products = cr.fetchall()
    logger.debug(
        "Migrating the valuation field of %s products with non-default values",
        len(products))

    seen_ids = []
    for template_id, value in products:
        if not value or template_id in seen_ids:
            continue
        seen_ids.append(template_id)
        property_obj.create(
            cr, SUPERUSER_ID, {
                'fields_id': field_id,
                'company_id': False,
                'res_id': 'product.template,{}'.format(template_id),
                'name': 'Valuation Property',
                'type': 'selection',
                'value_text': value,
                })

    cr.execute(
        "ALTER TABLE product_product DROP COLUMN {}".format(valuation_column))


def migrate_procurement_order_method(cr, pool):
    """Procurements method: change the supply_method for the matching rule

    Needs to be deferred because the rules are created in the migration
    of stock, purchase and mrp.

    Will only run if stock is installed. Will run after every attempt to
    upgrade, but harmless when run multiple times.
    """

    cr.execute(
        """
        SELECT id FROM ir_module_module
        WHERE name = 'stock' AND state = 'installed'
        AND latest_version = '8.0.1.1'
        """)
    if not cr.fetchone():
        # Only warn if there are traces of stock
        if openupgrade.table_exists(cr, 'stock_move'):
            logger.debug(
                "Stock not installed or not properly migrated, skipping "
                "migration of procurement orders.")
        return

    procure_method_legacy = openupgrade.get_legacy_name('procure_method')
    if not openupgrade.column_exists(
            cr, 'product_template', procure_method_legacy):
        # in this case, there was no migration for the procurement module
        # which can be okay if procurement was not installed in the 7.0 db
        return

    procurement_obj = pool['procurement.order']
    rule_obj = pool['procurement.rule']
    rules = {}
    for rule in rule_obj.browse(
            cr, SUPERUSER_ID,
            rule_obj.search(cr, SUPERUSER_ID, [])):
        rules.setdefault(
            rule.location_id.id, {})[rule.procure_method] = rule.id
        rules.setdefault(rule.location_id.id, {})[rule.action] = rule.id

    cr.execute(
        """
        SELECT pp.id FROM product_product pp, product_template pt
        WHERE pp.product_tmpl_id = pt.id
            AND pt.%s = 'produce'
        """ % openupgrade.get_legacy_name('supply_method'))
    production_products = [row[0] for row in cr.fetchall()]

    cr.execute(
        """
        SELECT id, %s FROM procurement_order
        WHERE rule_id is NULL AND state != %%s
        """ % procure_method_legacy, ('done',))

    procurements = cr.fetchall()
    if len(procurements):
        logger.debug(
            "Trying to find rules for %s procurements", len(procurements))

    for proc_id, procure_method in procurements:
        procurement = procurement_obj.browse(cr, SUPERUSER_ID, proc_id)
        location_id = procurement.location_id.id

        # if location type is internal (presumably stock), then
        # find the rule with this location and the appropriate action,
        # regardless of procure method
        rule_id = False
        action = 'move'  # Default, only for log message
        if procure_method == 'make_to_order':
            if procurement.location_id.usage == 'internal':
                if procurement.product_id.id in production_products:
                    action = 'manufacture'
                else:
                    action = 'buy'
                rule_id = rules.get(location_id, {}).get(action)
            else:
                rule_id = rules.get(location_id, {}).get('make_to_order')
        else:
            rule_id = rules.get(location_id, {}).get('make_to_stock')
        if rule_id:
            procurement.write({'rule_id': rule_id})
        else:
            logger.warn(
                "Procurement order #%s with location %s "
                "has no %s procurement rule with action %s, please create and "
                "assign a new rule for this procurement""",
                procurement.id, procurement.location_id.name,
                procure_method, action)


def migrate_stock_move_warehouse(cr):
    """
    If a database featured multiple shops with the same company but a
    different warehouse, we can now propagate this warehouse to the
    associated stock moves. The warehouses were written on the procurements
    in the sale_stock module, while the moves were associated with the
    procurements in purchase and mrp. The order of processing between
    these modules seems to be independent, which is why we do this here
    in the deferred step.
    """
    cr.execute(
        "SELECT * FROM ir_module_module WHERE name='stock' "
        "AND state='installed'")
    if not cr.fetchone():  # No stock
        return
    openupgrade.logged_query(
        cr,
        """
        UPDATE stock_move sm
        SET warehouse_id = po.warehouse_id
        FROM procurement_order po
        WHERE sm.procurement_id = po.id
            OR po.move_dest_id = sm.id
        """)


def migrate_deferred(cr, pool):
    migrate_product_valuation(cr, pool)
    migrate_procurement_order_method(cr, pool)
    migrate_stock_move_warehouse(cr)
